import logging

from crjbsim import time_provider

logger = logging.getLogger(__name__)


class EventQueue:
    def __init__(self):
        self._events: set[Event] = set()

    def pop_next(self):
        # TODO: change to store sorted
        next_event = min(self._events, key=lambda event: event.time)
        self._events.remove(next_event)
        return next_event

    def add(self, event):
        self._events.add(event)

    def __bool__(self):
        return bool(self._events)

    def __len__(self):
        return len(self._events)


class Event:
    def __init__(self, time, runnable):
        self.time = time
        self.runnable = runnable

    def execute(self):
        self.runnable()

    def __repr__(self):
        return f"Event({self.time}, {self.runnable})"


class DiscreteEventScheduler:
    def __init__(self):
        self._events = EventQueue()

    def start(self):
        while self._events:
            event = self._events.pop_next()
            assert event.time >= time_provider.get_time()
            time_provider.set_time(event.time)
            logger.debug(f"Executing event {event}")
            event.execute()

    def do_at(self, time, runnable):
        self._events.add(Event(time, runnable))

    def do_in(self, time_delta, runnable):
        self._events.add(Event(time_provider.get_time() + time_delta, runnable))



import random

from crjbsim import des_aware_logging
from crjbsim.discrete_event_scheduler import DiscreteEventScheduler


scheduler = None


def run_sim(initial_event, post_processing=None):
    random.seed(0)
    des_aware_logging.setup()

    global scheduler
    scheduler = DiscreteEventScheduler()

    scheduler.do_in(0, initial_event)

    scheduler.start()

    if post_processing:
        post_processing()



from crjbsim import sim


def test_sim():
    test_list = []
    sim.run_sim(lambda: test_list.append(1))
    assert test_list == [1]


def test_sim_post_process():
    test_list = []
    sim.run_sim(lambda: test_list.append(1), lambda: test_list.append(2))
    assert test_list == [1, 2]


def test_sim_scheduling():
    test_list = []
    sim.run_sim(lambda: sim.scheduler.do_in(5, lambda: test_list.append(1)))
    assert test_list == [1]



import logging
from logging import StreamHandler

from crjbsim import time_provider


class Handler(StreamHandler):
    def format(self, record):
        return f"{time_provider.get_time_formatted()} {super().format(record)}"


def setup():
    root_logger = logging.getLogger()
    handler = Handler()
    handler.formatter = logging.Formatter('%(levelname)-8s: %(message)s (%(name)s)')
    root_logger.handlers = [handler]
    root_logger.setLevel(logging.DEBUG)



import datetime


class _TimeProvider:
    def __init__(self):
        self.time = 0  # ms


_instance = _TimeProvider()


def get_time() -> int:
    """Gets the system time in milliseconds"""
    return _instance.time


def get_time_formatted() -> str:
    return datetime.datetime.fromtimestamp(_instance.time / 1000).strftime('%H:%M:%S.%f')[:-3]


def set_time(time: int):
    """Sets the system time in milliseconds. Should only be called from the discrete event scheduler"""
    _instance.time = time
